{
  Original Name: Twitter.pas
  Copyright: Copyright (c) 2010, Simon J Stuart, All Rights Reserved
  License: Licensed under the MIT License.
  Author: Simon J Stuart
  Author's Website: http://www.lakraven.com
  Description:
    This unit contains the TTwitter component, an easy-to-use
    layer which performs all interactions with the Twitter API
    for you, leaving you with standardized Delphi classes that
    contain the required information in a friendly format.
  Author's Notes:
    You are encouraged to include a credit somewhere within any
    application(s) using this unit (and those related) resembling
    the following:
    "This application uses TTwitter, produced by Simon J Stuart
    whose website can be found at http://www.lakraven.com"
    Whilst your use of this unit (and those related) is not
    contingent on provision of a notice or credit, it is appreciated.
    I also appreicate "success stories" or comments related to my
    code. You can find details of how to contact me at my website
    (see "Author's Website" for URL).
  Acknowledgements:
    - Jim McKeeth       : Testing, code proofing, XE trials, useful suggestions,
                          being an all-round good guy!
                          (www.delphi.org)
    - Chuck Beasley     : Providing invaluable insight into OAuth with Twitter,
                          all in the necessary Delphi context.
                          His work provided a solid foundation on which I was
                          able to build... even though he didn't know it!
                          (if you're reading this, Chuck, thanks my new friend!)
                          (www.chuckbeasley.com)
    - Twitter API Staff : Answering many questions, even the stupid ones.
                          (www.twitter.com)

  Version: 0.1 OFFICIAL OPEN BETA RELEASE
  --------------------------------------------------------------------------------
  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
  THE SOFTWARE.
  --------------------------------------------------------------------------------
  CODE NOTES:
  - Should you encounter any errors, send details as a comment to:
    http://www.lakraven.com/delphi-stuff/ttwitter
    Please include the code which causes the error, as well as any Message(s)
    you see at the point of error. You can also contact me on Twitter: @LaKraven
  - Eventually a lot of Twitter actions will become Threaded to provide a smooth
    runtime experience in your application(s) for the user(s).
    At present everything executes in the same Thread (easier debugging)

  IMPORTANT NOTICE: You are currently looking at an Open BETA version of TTwitter!
                    This code is subject to a LOT of changes, and whilst you may
                    find portions of this code at present to be messy... rest
                    assured that upon final release, it'll all be nice and clean!
}

unit Twitter;

interface

{$I Twitter.inc} // TTwitter's Configuration Data

uses
  Windows, Classes, SysUtils,
  Twitter_Common, Twitter_StringComposer, Twitter_OAuth, Twitter_OLogin,
  Twitter_Data, Twitter_TokenStore, Twitter_HTTP;

type
  {
    Forward Declaration of Classes
  }
  TTwitter = class;
  TTwitterUserManager = class;
  TTwitterTimelineManager = class;

  {
    Various "OnEvent" method types
  }
  TTwitterStatusEvent = procedure(const AResponseCode: Integer; const AMsg: String; const AAction: TTwitterStatusAction) of object;
  TTwitterExceptionEvent = procedure(const AMsg: String) of object; // For when TTwitter fails
  TTwitterUsernameChangeEvent = procedure(const ANewUsername: String) of object;
  TTwitterRawResponseEvent = procedure(const AURL, AResponse: String) of object;
  TTwitterRateLimitStatusEvent = procedure(const ARateLimitDetails: TTwitterRateLimitData) of object;

  {
    TTwitter object does all the magic! This is the actual component which registers on your
    Component Pallet.
    - All communications with Twitter are neatly exposed as native Delphi methods.
    - "Get" responses are provided to use by On<event> methods!
    - "Post" responses are returned as Function Results!
  }
  TTwitter = class(TComponent)
  private
    FApplicationDetails: TTwitterApplicationDetails;
    FAutoRateCheck: Boolean;
    FHTTP: TTwitterHTTP;
    FLoggedIn: Boolean;
    FParser: TTwitterParser;
    FProxy: TTwitterProxyDetails;
    FTimelines: TTwitterTimelineManager;
    FTokenManager: TTwitterTokenManager;
    FURLFactory: TTwitterURLAssembler;
    FUsername: String;
    FUsers: TTwitterUserManager;
    FUserToken: TTwitterUserToken;
    FUseXAuth: Boolean;
    FOnAsyncTimeline: TTwitterAsyncTimelineEvent;
    FOnError: TTwitterStatusEvent;
    FOnException: TTwitterExceptionEvent;
    FOnRateLimitStatus: TTwitterRateLimitStatusEvent;
    FOnRawResponse: TTwitterRawResponseEvent; // DESIGNED ONLY TO BE USED FOR DEBUGGING PURPOSES DURING TTWITTER'S OWN DEVELOPMENT!
    FOnUsernameChange: TTwitterUsernameChangeEvent;
    function GetDefaultMessage(const AResponseCode: Integer): String;
    function TwitterGET(const ARequiresAuth: Boolean; var AURL: String; const ABetterWithAuth: Boolean = True): String;
    function TwitterPOST(var AURL: String; var AParams: TTwitterParamFactory; MultiPart: Boolean = False): String;
    procedure DoStatus(const AResponseCode: Integer; const AMsg: String; const AAction: TTwitterStatusAction; const AStatusType: TTwitterStatusType);
    function GetUserInfo(const AUsername: String; const AIncludeEntities: Boolean = True): TTwitterUser; overload;
    function GetUserInfo(const AID: Int64; const AIncludeEntities: Boolean = True): TTwitterUser; overload;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function Login: Boolean;
    procedure ForgetUser; // will remove any Stored Token/Secret for the current/named User
    procedure GetRateLimitStatus;
    procedure Logout;
    function PostTweet(const AMsg: String; const AReplyToStatusID: Int64 = 0; const AImage: string = ''; const ALat: Extended = 181; const ALong: Extended = 181; const APlaceID: String = ''; const ADisplayCoordinates: Boolean = False; ATrimUser: Boolean = False; AIncludeEntities: Boolean = True): TTwitterTweet;
  published
    property ApplicationDetails: TTwitterApplicationDetails read FApplicationDetails write FApplicationDetails;
    property AutoRateCheck: Boolean read FAutoRateCheck write FAutoRateCheck default False;
    property LoggedIn: Boolean read FLoggedIn;
    property Proxy: TTwitterProxyDetails read FProxy write FProxy;
    property Timelines: TTwitterTimelineManager read FTimelines write FTimelines;
    property TokenStore: TTwitterTokenManager read FTokenManager write FTokenManager;
    property Username: String read FUsername write FUsername;
    property Users: TTwitterUserManager read FUsers write FUsers;
    property UseXAuth: Boolean read FUseXAuth write FUseXAuth default False;
    { OnEvent Methods }
    property OnAsyncTimeline: TTwitterAsyncTimelineEvent read FOnAsyncTimeline write FOnAsyncTimeline;
    property OnError: TTwitterStatusEvent read FOnError write FOnError;
    property OnException: TTwitterExceptionEvent read FOnException write FOnException;
    property OnRateLimitStatus: TTwitterRateLimitStatusEvent read FOnRateLimitStatus write FOnRateLimitStatus;
    property OnRawResponse: TTwitterRawResponseEvent read FOnRawResponse write FOnRawResponse;
    property OnUsernameChange: TTwitterUsernameChangeEvent read FOnUsernameChange write FOnUsernameChange;
  end;

  {
    TTwitterUserManager object provides us a neat means of managing Twitter Users (details and actions).
    It can cache retrieved User Details to make future interrogation of those details quicker!
  }
  TTwitterUserManager = class(TPersistent)
  private
    FTwitter: TTwitter;
    FUsers: TTwitterUsers;
    FCache,
    FIncludeEntities: Boolean;
    function GetUser(const AUsername: String): TTwitterUser; // Get user by SCREEN NAME
    function GetUserByID(const AIndex: Int64): TTwitterUser; // Get user by ID
  public
    constructor Create(ATwitter: TTwitter);
    destructor Destroy; override;
    property User[const AUsername: String]: TTwitterUser read GetUser; default;
    property User[const AIndex: Int64]: TTwitterUser read GetUserByID; default;
  published
    property Cache: Boolean read FCache write FCache default True;
    property IncludeEntities: Boolean read FIncludeEntities write FIncludeEntities default True;
  end;

  {
    TTwitterTimeline object provides us with multiple Tweets, as well as operations we can
    perform on them as appropriate.
  }
  TTwitterTimeline = class(TPersistent)
  private
    FCache: Boolean;
    FMaxStore: Integer;
    FTimelineManager: TTwitterTimelineManager;
    FTimelineType: TTwitterTimelines;
    FSinceID,
    FMaxID: Int64;
    FNumberToGet,
    FPage: Integer;
    FIncludeEntities,
    FTrimUser,
    FRetweets: Boolean;
    FScreenName: String;
    FUserID: Int64;
    FTotal: Integer;
    FTweets: TTweetArray;
    function GetTweetByIndex(const AIndex: Integer): TTwitterTweet;
    function GetTweetByID(const AID: Int64): TTwitterTweet;
    procedure AddTweet(const ATimeline: TTwitterTimelines; const ATweet: TTwitterTweet);
  public
    constructor Create(const ATimelineManager: TTwitterTimelineManager; const ATimeline: TTwitterTimelines); virtual;
    destructor Destroy; override;
    procedure GetTimeline;
    procedure Clear;
    property TweetByIndex[const AIndex: Integer]: TTwitterTweet read GetTweetByIndex; default;
    property TweetByID[const AID: Int64]: TTwitterTweet read GetTweetByID;
  published
    property TweetCount: Integer read FTotal write FTotal default 0;
    property Cache: Boolean read FCache write FCache default True;
    property MaxStore: Integer read FMaxStore write FMaxStore default 0;
  end;

  {
    TTwitterFriendsTimeline is a specialized version of TTwitterTimeline with additional
    properties/methods as required.
  }
  TTwitterFriendsTimeline = class(TTwitterTimeline)
  private
    FNumberToGet,
    FPage: Integer;
    FSinceID,
    FMaxID: Int64;
    FIncludeEntities,
    FTrimUser,
    FRetweets: Boolean;
  published
    property NumberToGet: Integer read FNumberToGet write FNumberToGet default 0;
    property Page: Integer read FPage write FPage default 0;
    property SinceID: Int64 read FSinceID write FSinceID default 0;
    property MaxID: Int64 read FMaxID write FMaxID default 0;
    property IncludeEntities: Boolean read FIncludeEntities write FIncludeEntities default True;
    property TrimUser: Boolean read FTrimUser write FTrimUser default True;
    property Retweets: Boolean read FRetweets write FRetweets default True;
  end;

  {
    TTwitterHomeTimeline is a specialized version of TTwitterTimeline with additional
    properties/methods as required.
  }
  TTwitterHomeTimeline = class(TTwitterTimeline)
  private
    FNumberToGet,
    FPage: Integer;
    FSinceID,
    FMaxID: Int64;
    FIncludeEntities,
    FTrimUser: Boolean;
  published
    property NumberToGet: Integer read FNumberToGet write FNumberToGet default 0;
    property Page: Integer read FPage write FPage default 0;
    property SinceID: Int64 read FSinceID write FSinceID default 0;
    property MaxID: Int64 read FMaxID write FMaxID default 0;
    property IncludeEntities: Boolean read FIncludeEntities write FIncludeEntities default True;
    property TrimUser: Boolean read FTrimUser write FTrimUser default True;
  end;

  {
    TTwitterMentionsTimeline is a specialized version of TTwitterTimeline with additional
    properties/methods as required.
  }
  TTwitterMentionsTimeline = class(TTwitterTimeline)
  private
    FNumberToGet,
    FPage: Integer;
    FSinceID,
    FMaxID: Int64;
    FIncludeEntities,
    FTrimUser,
    FRetweets: Boolean;
  published
    property NumberToGet: Integer read FNumberToGet write FNumberToGet default 0;
    property Page: Integer read FPage write FPage default 0;
    property SinceID: Int64 read FSinceID write FSinceID default 0;
    property MaxID: Int64 read FMaxID write FMaxID default 0;
    property IncludeEntities: Boolean read FIncludeEntities write FIncludeEntities default True;
    property TrimUser: Boolean read FTrimUser write FTrimUser default True;
    property Retweets: Boolean read FRetweets write FRetweets default True;
  end;

  {
    TTwitterPublicTimeline is a specialized version of TTwitterTimeline with additional
    properties/methods as required.
  }
  TTwitterPublicTimeline = class(TTwitterTimeline)
  private
    FIncludeEntities,
    FTrimUser: Boolean;
  published
    property IncludeEntities: Boolean read FIncludeEntities write FIncludeEntities default True;
    property TrimUser: Boolean read FTrimUser write FTrimUser default True;
  end;

  {
    TTwitterRetweetTimeline is a specialized version of TTwitterTimeline with additional
    properties/methods as required.
  }
  TTwitterRetweetTimeline = class(TTwitterTimeline)
  private
    FNumberToGet,
    FPage: Integer;
    FSinceID,
    FMaxID: Int64;
    FIncludeEntities,
    FTrimUser: Boolean;
  published
    property NumberToGet: Integer read FNumberToGet write FNumberToGet default 0;
    property Page: Integer read FPage write FPage default 0;
    property SinceID: Int64 read FSinceID write FSinceID default 0;
    property MaxID: Int64 read FMaxID write FMaxID default 0;
    property IncludeEntities: Boolean read FIncludeEntities write FIncludeEntities default True;
    property TrimUser: Boolean read FTrimUser write FTrimUser default True;
  end;

  {
    TTwitterFriendsTimeline is a specialized version of TTwitterTimeline with additional
    properties/methods as required.
  }
  TTwitterUserTimeline = class(TTwitterTimeline)
  private
    FNumberToGet,
    FPage: Integer;
    FSinceID,
    FMaxID: Int64;
    FIncludeEntities,
    FTrimUser,
    FRetweets: Boolean;
    FScreenName: String;
    FUserID: Int64;
  published
    property NumberToGet: Integer read FNumberToGet write FNumberToGet default 0;
    property Page: Integer read FPage write FPage default 0;
    property SinceID: Int64 read FSinceID write FSinceID default 0;
    property MaxID: Int64 read FMaxID write FMaxID default 0;
    property IncludeEntities: Boolean read FIncludeEntities write FIncludeEntities default True;
    property TrimUser: Boolean read FTrimUser write FTrimUser default True;
    property Retweets: Boolean read FRetweets write FRetweets default True;
    property ScreenName: String read FScreenName write FScreenName;
    property UserID: Int64 read FUserID write FUserID default 0;
  end;
  TTwitterUserTimelineArray = Array of TTwitterUserTimeline;
  {
    TTwitterMultiTimeline object is special as it provides us with a managed array of
    timelines. This is designed to apply to Users managed by Name and/or ID.
  }
  TTwitterMultiTimeline = class(TPersistent)
  private
    FCache: Boolean;
    FTimelines: TTwitterUserTimelineArray;
    FTimelineType: TTwitterTimelines;
    FTimelineManager: TTwitterTimelineManager;
    function GetTimeline(const AName: String): TTwitterUserTimeline; overload;
    function GetTimeline(const AIndex: Int64): TTwitterUserTimeline; overload;
  public
    constructor Create(const ATimelineManager: TTwitterTimelineManager; const ATimeline: TTwitterTimelines);
    destructor Destroy; override;
    property Timeline[const AName: String]: TTwitterUserTimeline read GetTimeline; default;
    property Timeline[const AIndex: Int64]: TTwitterUserTimeline read GetTimeline; default;
  published
    property Cache: Boolean read FCache write FCache default True;
  end;

  {
    TTwitterTimelineManager is a special base component providing us a means of
    accessing and managing the various Twitter Timelines.
  }
  TTwitterTimelineManager = class(TPersistent)
  private
    FPublic: TTwitterPublicTimeline;
    FHome: TTwitterHomeTimeline;
    FFriends: TTwitterFriendsTimeline;
    FMentions: TTwitterMentionsTimeline;
    FRetweetedByMe,
    FRetweetedToMe,
    FRetweetsOfMe: TTwitterRetweetTimeline;
    FTwitter: TTwitter;
    FUser: TTwitterMultiTimeline; // Special! One timeline per user!
    FNoDuplicates: Boolean;
  public
    constructor Create(const ATwitter: TTwitter); virtual;
    destructor Destroy; override;
  published
    property Public: TTwitterPublicTimeline read FPublic;
    property Home: TTwitterHomeTimeline read FHome;
    property Friends: TTwitterFriendsTimeline read FFriends;
    property Mentions: TTwitterMentionsTimeline read FMentions;
    property RetweetedByMe: TTwitterRetweetTimeline read FRetweetedByMe;
    property RetweetedToMe: TTwitterRetweetTimeline read FRetweetedToMe;
    property RetweetsOfMe: TTwitterRetweetTimeline read FRetweetsOfMe;
    property Users: TTwitterMultiTimeline read FUser; // Special! One timeline per user!
    property NoDuplicates: Boolean read FNoDuplicates write FNoDuplicates default True;
  end;

const
  TTwitter_Registry = '\Software\%s\Stored\'; // Default Registry location to store Tokens (if no custom instruction given)

implementation

{ TTwitter }

constructor TTwitter.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  if FApplicationDetails = nil then
    FApplicationDetails := TTwitterApplicationDetails.Create;
  FURLFactory := TTwitterURLAssembler.Create{(Self)};
  FParser := TTwitterParser.Create;
  FLoggedIn := False;
  FAutoRateCheck := False;
  FHTTP := TTwitterHTTP.Create(FApplicationDetails);
  FUsers := TTwitterUserManager.Create(Self);
  FProxy := TTwitterProxyDetails.Create;
  FTimelines := TTwitterTimelineManager.Create(Self);
end;

destructor TTwitter.Destroy;
begin
  FreeAndNil(FApplicationDetails);
  if FHTTP <> nil then
    FHTTP.Free;
  FURLFactory.Free;
  FParser.Free;
  FUsers.Free;
  FProxy.Free;
  FTimelines.Free;
  inherited Destroy;
end;

procedure TTwitter.DoStatus(const AResponseCode: Integer; const AMsg: String; const AAction: TTwitterStatusAction; const AStatusType: TTwitterStatusType);
begin
  case AStatusType of
    stError:
      begin
        if Assigned(FOnError) then
          FOnError(AResponseCode, AMsg, AAction)
        else
          raise ETwitterException.Create(AMsg);
      end;
    stException:
      begin
        if Assigned(FOnException) then
          FOnException(AMsg)
        else
          raise ETwitterException.Create(AMsg);
      end;
  end;
end;

function TTwitter.GetDefaultMessage(const AResponseCode: Integer): String;
begin
  case AResponseCode of
    304: Result := TTwitter_Error_304;
    400: Result := TTwitter_Error_400;
    401: Result := TTwitter_Error_401;
    403: Result := TTwitter_Error_403;
    404: Result := TTwitter_Error_404;
    406: Result := TTwitter_Error_406;
    420: Result := TTwitter_Error_420;
    500: Result := TTwitter_Error_500;
    502: Result := TTwitter_Error_502;
    503: Result := TTwitter_Error_503;
  end;
end;

function TTwitter.TwitterGET(const ARequiresAuth: Boolean; var AURL: String; const ABetterWithAuth: Boolean = True): String;
begin
  Result := FHTTP.Get(AURL, ARequiresAuth, FProxy, ABetterWithAuth);
  if FAutoRateCheck then
    GetRateLimitStatus;
end;

function TTwitter.TwitterPOST(var AURL: String; var AParams: TTwitterParamFactory; MultiPart: Boolean): String;
begin
  Result := FHTTP.Post(AURL, AParams, FProxy, MultiPart);
  if FAutoRateCheck then
    GetRateLimitStatus
end;

procedure TTwitter.GetRateLimitStatus;
var
  URL: String;
  GETResponse: String;
  LRateLimit: TTwitterRateLimitData;
begin
  URL := FURLFactory.RateLimitStatus;

  // Get API Limits (if logged in)
  if FLoggedIn then
  begin
    GETResponse := FHTTP.Get(URL, False, FProxy);
    LRateLimit := FParser.ParseRateLimit(GETResponse);
  end else
  begin
    GETResponse := FHTTP.Get(URL, False, FProxy, False);
    LRateLimit := FParser.ParseRateLimit(GETResponse);
  end;

  if Assigned(FOnRateLimitStatus) then
    FOnRateLimitStatus(LRateLimit);
end;

function TTwitter.GetUserInfo(const AUsername: String; const AIncludeEntities: Boolean = True): TTwitterUser;
var
  LURL: String;
  GETResponse: String;
begin
  LURL := FURLFactory.UsersShow(AUsername, AIncludeEntities);
  GETResponse := TwitterGET(False, LURL);
  if Assigned(FOnRawResponse) then // TODO: Temporary (Debugging), Remove this!
    FOnRawResponse(LURL, GETResponse);
  Result := FParser.ParseUserInfo(GETResponse);
end;

function TTwitter.GetUserInfo(const AID: Int64; const AIncludeEntities: Boolean = True): TTwitterUser;
var
  LURL: String;
  GETResponse: String;
begin
  LURL := FURLFactory.UsersShow(AID, AIncludeEntities);
  GETResponse := TwitterGET(False, LURL);
  Result := FParser.ParseUserInfo(GETResponse);
end;

function TTwitter.Login: Boolean;
begin
  if FLoggedIn then
    Logout;

  FLoggedIn := FHTTP.Login(FUsername, FTokenManager, FUseXAuth, FProxy);

  Result := FLoggedIn;

  if Result then
  begin
    FUsername := FHTTP.UserToken.Username;
    if Assigned(FOnUsernameChange) then
      FOnUsernameChange(FHTTP.UserToken.Username);
  end;
end;

procedure TTwitter.Logout;
begin
  FHTTP.Logout;
  if FLoggedIn then
    FLoggedIn := False;
end;

procedure TTwitter.ForgetUser;
begin
  if Assigned(FTokenManager) then
    FTokenManager.KillToken(FUserToken);
end;

function TTwitter.PostTweet(const AMsg: String; const AReplyToStatusID: Int64; const AImage: string; const ALat, ALong: Extended; const APlaceID: String; const ADisplayCoordinates: Boolean; ATrimUser, AIncludeEntities: Boolean): TTwitterTweet;
var
  LParams: TTwitterParamFactory;
  URL,
  GETResponse: String;
begin
  Result := nil;
  LParams := TTwitterParamFactory.Create;
  try
    if (Length(AMsg) = 0) or (Length(AMsg) > 140) then
    begin
      DoStatus(0, TTwitter_Error_StatusLength, saPostTweet, stError);
      exit;
    end;

    URL := FURLFactory.PostTweet(AMsg, AReplyToStatusID, AImage, ALat, ALong, APlaceID, ADisplayCoordinates, ATrimUser, AIncludeEntities, LParams);
    try
      GETResponse := TwitterPOST(URL, LParams, (AImage <> '') and FileExists(AImage));
      Result := FParser.ParseTweet(GETResponse);
    except
      // don't complain that this exception handler somehow "hides" the exception, it doesn't! It simply
      // displays any exceptions using TTwitter's own custom Exception system!
      on E: Exception do
      begin
        if (FHTTP.HTTP.ResponseCode <> 200) and (FHTTP.HTTP.ResponseCode > 300) then
          DoStatus(FHTTP.HTTP.ResponseCode, E.Message, saPostTweet, stException)
        else
          DoStatus(FHTTP.HTTP.ResponseCode, GetDefaultMessage(FHTTP.HTTP.ResponseCode), saPostTweet, stError);
        exit;
      end;
    end;
    if FHTTP.HTTP.ResponseCode <> 200 then
    begin
      // Error Response given by Twitter's API...
      DoStatus(FHTTP.HTTP.ResponseCode, GetDefaultMessage(FHTTP.HTTP.ResponseCode), saPostTweet, stError);
      exit;
    end;
  finally
    LParams.Free;
  end;
end;

{ TTwitterUserManager }

constructor TTwitterUserManager.Create(ATwitter: TTwitter);
begin
  inherited Create;
  FCache := True;
  FIncludeEntities := True;
  FTwitter := ATwitter;
end;

destructor TTwitterUserManager.Destroy;
var
  I: Integer;
begin
  for I := 0 to Length(FUsers) - 1 do
  begin
    if FUsers[I] <> nil then
      FreeAndNil(FUsers[I]);
  end;
  SetLength(FUsers, 0);
  inherited;
end;

function TTwitterUserManager.GetUser(const AUsername: String): TTwitterUser;
var
  LIdx: Integer;
begin
  Result := nil;
  if FCache then
  begin
    for LIdx := 0 to Length(FUsers) - 1 do
    begin
      if FUsers[LIdx] <> nil then
      begin
        if FUsers[LIdx].ScreenName =  AUsername then
        begin
          Result := FUsers[LIdx];
          Exit;
        end;
      end;
    end;
  end;
  if Assigned(FTwitter) then
  begin
    LIdx := Length(FUsers);
    SetLength(FUsers, LIdx + 1);
    FUsers[LIdx] := FTwitter.GetUserInfo(AUsername, FIncludeEntities);
    Result := FUsers[LIdx];
  end;
end;

function TTwitterUserManager.GetUserByID(const AIndex: Int64): TTwitterUser;
var
  LIdx: Integer;
begin
  Result := nil;
  if FCache then
  begin
    for LIdx := 0 to Length(FUsers) - 1 do
    begin
      if FUsers[LIdx] <> nil then
      begin
        if FUsers[LIdx].ID =  AIndex then
        begin
          Result := FUsers[LIdx];
          Exit;
        end;
      end;
    end;
  end;
  if Assigned(FTwitter) then
  begin
    LIdx := Length(FUsers);
    SetLength(FUsers, LIdx + 1);
    FUsers[LIdx] := FTwitter.GetUserInfo(AIndex, FIncludeEntities);
    Result := FUsers[LIdx];
  end;
end;

{ TTwitterTimeline }

procedure TTwitterTimeline.AddTweet(const ATimeline: TTwitterTimelines; const ATweet: TTwitterTweet);
var
  LIdx: Integer;
  LTweetUnique: Boolean;
begin
  if ATweet = nil then
    Exit;
// TODO: Figure out why "since_id" parameter results in 401 errors
//  if ATweet.ID > FSinceID then
//    FSinceId := ATweet.ID;

  LTweetUnique := True;

  if Self.GetTweetByID(ATweet.ID) = nil then
  begin
    LIdx := Length(FTweets);
    SetLength(FTweets, LIdx + 1);
    FTweets[LIdx] := ATweet;
    FTotal := LIdx + 1;
  end else
    LTweetUnique := False;

  if Assigned(FTimelineManager.FTwitter.FOnAsyncTimeline) then
  begin
    if ((not FTimelineManager.FNoDuplicates) and (not LTweetUnique)) or (LTweetUnique)then
      FTimelineManager.FTwitter.FOnAsyncTimeline(ATimeline, ATweet);
  end;

  if (not LTweetUnique) then
    ATweet.Free;
end;

procedure TTwitterTimeline.Clear;
var
  I: Integer;
begin
  for I := 0 to FTotal -1 do
    FTweets[I].Free;
  SetLength(FTweets, 0);
end;

constructor TTwitterTimeline.Create(const ATimelineManager: TTwitterTimelineManager; const ATimeline: TTwitterTimelines);
begin
  FTimelineManager := ATimelineManager;
  FTimelineType := ATimeline;
  FCache := True;
  FMaxStore := 0;
  FNumberToGet := 0;
  FSinceID := 0;
  FMaxID := 0;
  FPage := 0;
  FIncludeEntities := True;
  FTrimUser := False;
  FRetweets := True;
  FScreenName := EmptyStr;
  FUserID := 0;
  FTotal := 0;
end;

destructor TTwitterTimeline.Destroy;
begin
  Clear;
  inherited;
end;

procedure TTwitterTimeline.GetTimeline;
var
  LURL, GETResponse: String;
begin
  if (not Assigned(FTimelineManager.FTwitter)) then
    Exit;
  if FTimelineManager.FTwitter = nil then
    Exit;
  if FTimelineType = ttNone then
    Exit; // TODO: Exception Handler

  LURL := FTimelineManager.FTwitter.FURLFactory.Timeline(FTimelineType, FNumberToGet, FSinceID, FMaxID, FPage, FScreenName, FIncludeEntities, FTrimUser, FRetweets, FUserID);

  GETResponse := FTimelineManager.FTwitter.TwitterGET((FTimelineType <> ttUser) and (FTimelineType <> ttPublic), LURL);
  // TODO: Exception Handler

  FTimelineManager.FTwitter.FParser.ParseTimeline(GETResponse, AddTweet, FTimelineType);
end;

function TTwitterTimeline.GetTweetByID(const AID: Int64): TTwitterTweet;
var
  I: Integer;
begin
  Result := nil;
  for I := 0 to FTotal - 1 do
  begin
    if FTweets[I].ID = AID then
    begin
      Result := FTweets[I];
      Break;
    end;
  end;
end;

function TTwitterTimeline.GetTweetByIndex(const AIndex: Integer): TTwitterTweet;
begin
  if (AIndex >= 0) and (AIndex <= FTotal) then
    Result := FTweets[AIndex]
  else
    Result := nil;
end;

{ TTwitterMultiTimeline }

function TTwitterMultiTimeline.GetTimeline(const AName: String): TTwitterUserTimeline;
begin

end;

function TTwitterMultiTimeline.GetTimeline(const AIndex: Int64): TTwitterUserTimeline;
begin

end;

constructor TTwitterMultiTimeline.Create(const ATimelineManager: TTwitterTimelineManager; const ATimeline: TTwitterTimelines);
begin
  inherited Create;
  FTimelineManager := ATimelineManager;
  FTimelineType := ATimeline;
end;

destructor TTwitterMultiTimeline.Destroy;
var
  I: Integer;
begin
  for I := 0 to Length(FTimelines) - 1 do
    FTimelines[I].Free;
  inherited;
end;

{ TTwitterTimelineManager }

constructor TTwitterTimelineManager.Create(const ATwitter: TTwitter);
begin
  inherited Create;
  FPublic := TTwitterPublicTimeline.Create(Self, ttPublic);
  FHome := TTwitterHomeTimeline.Create(Self, ttHome);
  FFriends := TTwitterFriendsTimeline.Create(Self, ttFriends);
  FMentions := TTwitterMentionsTimeline.Create(Self, ttMentions);
  FRetweetedByMe := TTwitterRetweetTimeline.Create(Self, ttRetweetedByMe);
  FRetweetedToMe := TTwitterRetweetTimeline.Create(Self, ttRetweetedToMe);
  FRetweetsOfMe := TTwitterRetweetTimeline.Create(Self, ttRetweetsOfMe);
  FUser := TTwitterMultiTimeline.Create(Self, ttUser);
  FTwitter := ATwitter;
  FNoDuplicates := True;
end;

destructor TTwitterTimelineManager.Destroy;
begin
  FPublic.Free;
  FHome.Free;
  FFriends.Free;
  FMentions.Free;
  FRetweetedByMe.Free;
  FRetweetedToMe.Free;
  FRetweetsOfMe.Free;
  FUser.Free;
  inherited;
end;

end.
